// Copyright The pipewire-rs Contributors.
// SPDX-License-Identifier: MIT

use std::{ffi::c_void, mem, pin::Pin};

use spa::spa_interface_call_method;

use crate::{
    proxy::{Listener, Proxy, ProxyT},
    types::ObjectType,
};

#[derive(Debug)]
pub struct Profiler {
    proxy: Proxy,
}

impl ProxyT for Profiler {
    fn type_() -> ObjectType {
        ObjectType::Profiler
    }

    fn upcast(self) -> Proxy {
        self.proxy
    }

    fn upcast_ref(&self) -> &Proxy {
        &self.proxy
    }

    unsafe fn from_proxy_unchecked(proxy: Proxy) -> Self
    where
        Self: Sized,
    {
        Self { proxy }
    }
}

impl Profiler {
    #[must_use]
    pub fn add_listener_local(&self) -> ProfilerListenerLocalBuilder {
        ProfilerListenerLocalBuilder {
            profiler: self,
            cbs: ListenerLocalCallbacks::default(),
        }
    }
}

pub struct ProfilerListener {
    // Need to stay allocated while the listener is registered
    #[allow(dead_code)]
    events: Pin<Box<pw_sys::pw_profiler_events>>,
    listener: Pin<Box<spa_sys::spa_hook>>,
    #[allow(dead_code)]
    data: Box<ListenerLocalCallbacks>,
}

impl Listener for ProfilerListener {}

impl Drop for ProfilerListener {
    fn drop(&mut self) {
        spa::utils::hook::remove(*self.listener);
    }
}

#[derive(Default)]
struct ListenerLocalCallbacks {
    #[allow(clippy::type_complexity)]
    profile: Option<Box<dyn Fn(&[u8])>>,
}

pub struct ProfilerListenerLocalBuilder<'profiler> {
    profiler: &'profiler Profiler,
    cbs: ListenerLocalCallbacks,
}

impl<'a> ProfilerListenerLocalBuilder<'a> {
    #[must_use]
    pub fn profile<F>(mut self, profile: F) -> Self
    where
        F: Fn(&[u8]) + 'static,
    {
        self.cbs.profile = Some(Box::new(profile));
        self
    }

    #[must_use]
    pub fn register(self) -> ProfilerListener {
        unsafe extern "C" fn profiler_events_profile(
            data: *mut c_void,
            pod: *const spa_sys::spa_pod,
        ) {
            let callbacks = (data as *mut ListenerLocalCallbacks).as_ref().unwrap();

            let pod_slice = if !pod.is_null() {
                std::slice::from_raw_parts(
                    pod as *const u8,
                    (*pod).size as usize + 2 * std::mem::size_of::<u32>(),
                )
            } else {
                &[]
            };

            callbacks.profile.as_ref().unwrap()(pod_slice);
        }

        let e = unsafe {
            let mut e: Pin<Box<pw_sys::pw_profiler_events>> = Box::pin(mem::zeroed());
            e.version = pw_sys::PW_VERSION_PROFILER_EVENTS;

            if self.cbs.profile.is_some() {
                e.profile = Some(profiler_events_profile);
            }

            e
        };

        let (listener, data) = unsafe {
            let profiler = &self.profiler.proxy.as_ptr();

            let data = Box::into_raw(Box::new(self.cbs));
            let mut listener: Pin<Box<spa_sys::spa_hook>> = Box::pin(mem::zeroed());
            let listener_ptr: *mut spa_sys::spa_hook = listener.as_mut().get_unchecked_mut();

            spa_interface_call_method!(
                profiler,
                pw_sys::pw_profiler_methods,
                add_listener,
                listener_ptr.cast(),
                e.as_ref().get_ref(),
                data as *mut _
            );

            (listener, Box::from_raw(data))
        };

        ProfilerListener {
            events: e,
            listener,
            data,
        }
    }
}
